/* avagl: surface_profile.cc
 * 
 * Copyright 2003 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "surface_profile.hh"
#include <cmath>

SurfaceProfile::SurfaceProfile(unsigned w, unsigned l, double s) :
	Gtk::GLArea()
{
	yangle = 0;
	wid = w;
	len = l;
	scale = s;
	prof = NULL;
	change = false;

	set_events(GDK_EXPOSURE_MASK |
		   GDK_BUTTON_MOTION_MASK |
		   GDK_POINTER_MOTION_HINT_MASK |
		   GDK_BUTTON_PRESS_MASK |
		   GDK_BUTTON_RELEASE_MASK);
}

SurfaceProfile::~SurfaceProfile()
{
	if (glIsList(profile)) {
		glDeleteLists(profile, 2);
	}
}

void SurfaceProfile::realize_impl()
{
	Gtk::GLArea::realize_impl();

	set_usize(256, 256);

	if (make_current()) {
		setupgl();
		profile = glGenLists(2);

		draw_profile();

		// Bottom reference plate
		glNewList(profile + 1, GL_COMPILE);
		glShadeModel(GL_FLAT);
		glFrontFace(GL_CCW);
		glBegin(GL_QUADS);
		glNormal3f(0, 1, 0);
		glColor3f(0.0, 0.0, 0.5);
		glVertex3f(1, -0.8, 1);
		glVertex3f(1, -0.8, -1);
		glVertex3f(-1, -0.8, -1);
		glVertex3f(-1, -0.8, 1);
		glEnd();
		glEndList();
	}
}

gint SurfaceProfile::motion_notify_event_impl(GdkEventMotion *e)
{
	double x;
	double y;
	double dx;
	double dy;
	GdkModifierType state;

	if (e->is_hint) {
		gint tx;
		gint ty;
		gdk_window_get_pointer(e->window, & tx, & ty, & state);
		x = tx;
		y = ty;
	} else {
		x = e->x;
		y = e->y;
		state = GdkModifierType(e->state);
	}
	if (! (state & GDK_BUTTON1_MASK)) return true;
	dx = x - bx;
	dy = y - by;
	bx = x;
	by = y;

	yangle += dx;

	while (yangle < 0) yangle += 360;
	while (yangle > 360) yangle -=360;
	redraw();
	return true;
}

gint SurfaceProfile::button_press_event_impl(GdkEventButton * e)
{
	bx = e->x;
	by = e->y;
	return true;
}

gint SurfaceProfile::button_release_event_impl(GdkEventButton * e)
{
	return true;
}

void SurfaceProfile::draw_profile()
{

	if (prof == NULL) return;

	glNewList(profile, GL_COMPILE);
	glShadeModel(GL_FLAT);

	for (unsigned x = 0; x < wid; x ++) {
		glFrontFace(GL_CCW);
		glBegin(GL_TRIANGLE_STRIP);
		glNormal3f(0, 1, 0);
		double p0 = prof[x];
		double p1 = prof[x];
		for (unsigned y = 0; y < len; y ++) {
			double v = prof[(x + wid - (y + 1) / 2)	% wid
					+ y * wid];
			bool ac = active[(x + wid - (y + 1) / 2) % wid
					 + y * wid];
					 
			if (y > 1) {
				double zd = (p0 - v) / 2;
				double xd = ((y % 2 == 1)
					     ? (p1 * 2 - p0 - v) :
					     (p0 + v - p1 * 2)) / 2;
				glNormal3f(- xd, 1, - zd);
			}
			double c = 0.3 + 0.7 * v / (10 + fabs(v));
			if (ac) glColor3f(c / 2 + 0.5, c / 2, c / 2);
			else glColor3f(c, c, c);
			glVertex3f((1 + x * 2.0 - (y % 2)) / wid - 1.0,
				   v * scale - 0.6, y * 2.0 / len - 1.0);
			p0 = p1;
			p1 = v;
		}
		glEnd();
		glFrontFace(GL_CW);
		glBegin(GL_TRIANGLE_STRIP);
		glNormal3f(0, 1, 0);
		p0 = prof[x];
		p1 = prof[x];
		for (unsigned y = 0; y < len; y ++) {
			double v = prof[(x + wid - y / 2) % wid + y * wid];
			bool ac = active[(x + wid - y / 2) % wid + y * wid];
			if (y > 1) {
				double zd = (p0 - v)  / 2;
				double xd = ((y % 2 == 0) ?
					     (p1 * 2 - p0 - v) :
					     (p0 + v - p1 * 2)) / 2;
				glNormal3f(- xd, 1, - zd);
			}
			double c = 0.3 + 0.7 * v / (10 + fabs(v));
			if (ac) glColor3f(c / 2 + 0.5, c / 2, c / 2);
			else glColor3f(c, c, c);
			glVertex3f((1 + x * 2.0 + (y % 2)) / wid - 1.0,
				   v * scale - 0.6, y * 2.0 / len - 1.0);
			p0 = p1;
			p1 = v;
		}
		glEnd();
	}
	glEndList();
	change = false;
}

gint SurfaceProfile::expose_event_impl(GdkEventExpose * e)
{
	if (e->count > 0) return 0;
	if (make_current()) {
		if (change) draw_profile();
		drawgl();
	}
	return true;
}

gint SurfaceProfile::configure_event_impl(GdkEventConfigure * e)
{
	if (make_current())
		setupgl();
	return true;
}


void SurfaceProfile::setupgl()
{

	static GLfloat pos[4] = {5.0, 10.0, 5.0, 0.0};

	glClearColor(1.0, 1.0, 1.0, 0.0);
	glLightfv(GL_LIGHT0, GL_POSITION, pos);
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);
	// glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	glViewport(0, 0, (GLint)(width()), (GLint)(height()));

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	if (width() > height()) {
		GLfloat w = (GLfloat) width() / (GLfloat) height();
		glFrustum(- w, w, - 1.0, 1.0, 5.0, 60.0);
	} else {
		GLfloat h = (GLfloat) height() / (GLfloat) width();
		glFrustum(- 1.0, 1.0, - h, h, 5.0, 60.0);
	}
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0, 0.0, -8.0);
}

void SurfaceProfile::drawgl()
{
	if (make_current()) {
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glPushMatrix();
		glRotatef(25, 1.0, 0.0, 0.0);
		glRotatef(yangle, 0.0, 1.0, 0.0);
		glRotatef(0, 0.0, 0.0, 1.0);
		glCallList(profile + 1);
		glCallList(profile);
		glPopMatrix();
	}
	swap_buffers();
}

void SurfaceProfile::redraw()
{
	if (make_current()) {
		draw_profile();
		drawgl();
	}
}

void SurfaceProfile::rotate_y(double d)
{
	yangle += d;
	yangle -= 360 * int(yangle / 360);
	// drawgl();
}

void SurfaceProfile::set_profile(const int * p)
{
	prof = p;
}

void SurfaceProfile::set_active(const bool * a)
{
	active = a;
}
